"""
hopeit.engine CLI (Command Line Interface) module

provides implementation for the following CLI commands:

    * **openapi** (hopeit_openapi): creation, diff and update openapi.json spec files
    * **server** (hopeit_server): tool for running a server instance
"""
