"""
hopeit.engine toolkit

Helper modules intended to be used by apps to solve common problems

    * **auth**: helpers to work with authentication and authorization mechanisms
"""

__all__ = ['auth']
